from aiohttp import (
    web,
)

from ..domain import (
    CannotInstantiateException,
    Microservice,
)
from ..domain.endpoint import (
    ConcreteEndpoint,
)
from ..exceptions import (
    NotFoundException,
)
from .router import (
    routes,
)


@routes.view("/microservices")
class EndpointView(web.View):
    async def get(self):
        try:
            verb = self.request.query["verb"]
            path = self.request.query["path"]
        except KeyError:
            raise web.HTTPBadRequest(text="Missing either 'verb' or 'path' query params")

        redis_client = self.request.app["db_client"]

        try:
            endpoint = ConcreteEndpoint(verb, path)
        except CannotInstantiateException:
            raise web.HTTPBadRequest(text="The given endpoint cannot have parts enclosed in brackets")

        try:
            microservice = await Microservice.find_by_endpoint(endpoint, redis_client)
        except NotFoundException:
            return web.HTTPNotFound()

        return web.json_response(microservice.to_json())



from pathlib import (
    Path,
)
from typing import (
    Optional,
)

import typer

from minos.api_gateway.common import (
    MinosConfig,
)

from .health_status import (
    HealthStatusCheckerService,
)
from .launchers import (
    EntrypointLauncher,
)
from .service import (
    DiscoveryService,
)

app = typer.Typer()


@app.command("start")
def start(
    file_path: Optional[Path] = typer.Argument(
        "config.yml", help="Discovery Service configuration file.", envvar="MINOS_API_GATEWAY_CONFIG_FILE_PATH"
    )
):  # pragma: no cover
    """Start Discovery services."""

    config = MinosConfig(file_path)

    services = (
        HealthStatusCheckerService(interval=120, delay=0, config=config),
        DiscoveryService(
            address=config.discovery.connection.host, port=config.discovery.connection.port, config=config
        ),
    )
    launcher = EntrypointLauncher(services=services)  # pragma: no cover
    launcher.launch()


@app.command("status")
def status():
    """Get the Discovery Service status."""
    raise NotImplementedError


@app.command("stop")
def stop():
    """Stop the Discovery Service."""
    raise NotImplementedError


def main():  # pragma: no cover
    """CLI's main function."""
    app()



from __future__ import (
    annotations,
)

from .exceptions import (
    CannotInstantiateException,
)


class PathPart:
    def __init__(self, name: str):
        self.name = name
        self.is_generic: bool = True if self.name.startswith("{") and self.name.endswith("}") else False


class Endpoint:
    def __init__(self, verb: str, path: str):
        self.verb = verb
        self.path: tuple[PathPart] = tuple(PathPart(path_part) for path_part in path.split("/"))

    @property
    def path_as_str(self) -> str:
        return "/".join([str(part.name) for part in self.path])


class ConcreteEndpoint(Endpoint):
    def __init__(self, verb: str, path: str):
        super().__init__(verb, path)
        for part in self.path:
            if part.is_generic:
                raise CannotInstantiateException


class GenericEndpoint(Endpoint):
    """Generic Endpoint class."""

    @classmethod
    def load_by_key(cls, endpoint_key: bytes) -> GenericEndpoint:
        """Build a new generic endpoint from key.

        :param endpoint_key: A key codified in bytes.
        :return: A ``GenericEndpoint``.
        """
        _, verb, path = endpoint_key.decode("utf-8").split(":", 2)
        return cls(verb, path)

    def matches(self, concrete_endpoint: ConcreteEndpoint) -> bool:
        if self.verb != concrete_endpoint.verb:
            return False

        if len(self.path) != len(concrete_endpoint.path):
            return False
        for path_part, concrete_path_part in zip(self.path, concrete_endpoint.path):
            if path_part.name != concrete_path_part.name and not path_part.is_generic:
                return False

        return True



import asyncio
import logging

from aiohttp import (
    web,
)
from aiomisc.service.aiohttp import (
    AIOHTTPService,
)

from minos.api_gateway.common import (
    MinosConfig,
)

from .database import (
    MinosRedisClient,
)
from .views import (
    routes,
)

logger = logging.getLogger(__name__)


class DiscoveryService(AIOHTTPService):
    """Discovery Service class."""

    def __init__(self, address: str, port: int, config: MinosConfig, graceful_stop_timeout: int = 5):
        self.config = config
        self.graceful_stop_timeout = graceful_stop_timeout
        logger.info(
            f"Address {address}"
            f"Port {port}"
        )
        super().__init__('0.0.0.0', '5567')

    async def create_application(self) -> web.Application:
        app = web.Application()
        app.router.add_routes(routes)

        app["db_client"] = MinosRedisClient(self.config)

        return app

    async def stop(self, exception: Exception = None) -> None:
        """

        :param exception:
        :return:
        """

        logger.info(
            f"Stopping Discovery Service gracefully "
            f"(waiting {self.graceful_stop_timeout} seconds for microservices unsubscription)..."
        )
        await asyncio.sleep(self.graceful_stop_timeout)
        await super().stop(exception)
